#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"

#define MSG_REPS 10
#define DEF_BUFFLEN 64
#define MTAG 99

int main(int argc, char *argv[]) {
  int rank, nprocs;
  double *mbuf, *gbuf; int len, reps, *rcounts;
  int i, v;
  double t1, t2, tbar;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  len = (argc > 1)? atoi(argv[1]): DEF_BUFFLEN;
  reps = (len < 100? 100: (len < 1000? 10: 1)) * MSG_REPS;
  reps = (argc > 2)? atoi(argv[2]): reps;

  mbuf = calloc(len*nprocs, sizeof(double));
  gbuf = calloc(len*nprocs, sizeof(double));
  rcounts = malloc(nprocs * sizeof(int));
  for (i=0; i < nprocs; i++)
    rcounts[i] = len;

  MPI_Barrier(MPI_COMM_WORLD);
  MPI_Barrier(MPI_COMM_WORLD);
  t1 = MPI_Wtime();
  MPI_Barrier(MPI_COMM_WORLD);
  tbar = MPI_Wtime() - t1;
  if (rank==0)
    printf("allGather/reduceScatter program: %d procs %d words %d reps; barrier t=%.2es\n", nprocs, len, reps, tbar);

  for (v = 0; v < 3; v++) {
    t1 = 0.0;
    for (i=0; i < reps+1; i++) {
      if (i==1) t1 = MPI_Wtime(); //exclude rep 0 for warm-up
      switch (v) {
      case 0:
	MPI_Allgather(mbuf, len, MPI_DOUBLE, gbuf, len, MPI_DOUBLE, 
		      MPI_COMM_WORLD);
	break;
      case 1: // insert MPI_Gather() / MPI_Bcast() equivalent code here

	break;
      case 2: // insert equivalent point-to-point code here
        // you can use MTAG for tag, MPI_STATUS_IGNORE for status

	break;
      }
    } //for(i...)
    MPI_Barrier(MPI_COMM_WORLD);
    t2 = MPI_Wtime() - t1 - tbar;

    if (rank==0 && t2 > 0.0)
      printf("all-gather     v%d: av t=%.2es bw=%.2eMB/s\n", 
	     v, t2/reps, 1.0E-6*nprocs*len*reps/t2);

    t1 = 0.0;
    for (i=0; i < reps+1; i++) {
      if (i==1) t1 = MPI_Wtime(); //exclude rep 0 for warm-up
      switch (v) {
      case 0:
	MPI_Reduce_scatter(gbuf, mbuf, rcounts, MPI_DOUBLE, MPI_SUM,	
			   MPI_COMM_WORLD);
	break;
      case 1: // insert MPI_Reduce() / MPI_Scatter() equivalent code here
	// use MPI_Reduce(rank==0? MPI_IN_PLACE: gbuf, gbuf, ...)

	break;
      case 2: // insert equivalent point-to-point code here
        // you can use MTAG for tag, MPI_STATUS_IGNORE for status
        // mbuf may be used as the receive buffer for the reduction

	break;
      }
    } //for(i...)
    MPI_Barrier(MPI_COMM_WORLD);
    t2 = MPI_Wtime() - t1 - tbar;
  
    if (rank==0 && t2 > 0.0)
      printf("reduce-scatter v%d: av t=%.2es bw=%.2eMB/s\n", 
	     v, t2/reps, 1.0E-6*nprocs*len*reps/t2);
  } //for(v...)

  free(mbuf);
  free(gbuf);
  MPI_Finalize();
  return 0;
} //main()
